package eu.ztsh.wymiana.service;

import eu.ztsh.wymiana.exception.ExchangeFailedException;
import eu.ztsh.wymiana.exception.InsufficientFundsException;
import eu.ztsh.wymiana.model.Currency;
import eu.ztsh.wymiana.model.User;
import eu.ztsh.wymiana.validation.InstanceValidator;
import eu.ztsh.wymiana.web.model.CurrencyExchangeRequest;
import lombok.RequiredArgsConstructor;
import org.springframework.stereotype.Service;

import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@RequiredArgsConstructor
@Service
public class CurrencyService {

    private final UserService userService;
    private final NbpService nbpService;
    private final InstanceValidator validator;

    public User exchange(CurrencyExchangeRequest request) {
        validator.validate(request);
        return userService.get(request.pesel()).map(user -> {
                if (!request.from().equalsIgnoreCase("PLN") && !request.to().equalsIgnoreCase("PLN")) {
                    throw new ExchangeFailedException("Either 'from' or 'to' has to be PLN");
                }
                // As we support only USD now, we need to limit second parameter too
                // Begin: unlock other currencies
                if (!request.from().equalsIgnoreCase("USD") && !request.to().equalsIgnoreCase("USD")) {
                    throw new ExchangeFailedException("Either 'from' or 'to' has to be USD");
                }
                // End: unlock other currencies

                var exchanged = performExchange(user.currencies().get(request.from().toUpperCase()),
                    user.currencies().get(request.to().toUpperCase()),
                    Optional.ofNullable(request.toSell()).orElse(0D),
                    Optional.ofNullable(request.toBuy()).orElse(0D));
                user.currencies().putAll(exchanged);
                return userService.update(user);
            })
            .orElseThrow(ExchangeFailedException::new);
    }

    private Map<String, Currency> performExchange(Currency from, Currency to, double toSell, double toBuy) {
        double exchangeRate;
        double neededFromAmount;
        double requestedToAmount;
        if (from.symbol().equalsIgnoreCase("PLN")) {
            exchangeRate = nbpService.getSellRate(to.symbol());
            neededFromAmount = toBuy != 0 ? toBuy * exchangeRate : toSell;
            requestedToAmount = toBuy != 0 ? toBuy : toSell / exchangeRate;
        } else {
            exchangeRate = nbpService.getSellRate(from.symbol());
            neededFromAmount = toBuy != 0 ? toBuy / exchangeRate : toSell;
            requestedToAmount = toBuy != 0 ? toBuy : toSell * exchangeRate;
        }
        if (neededFromAmount > from.amount()) {
            throw new InsufficientFundsException();
        }
        var newFrom = new Currency(from.symbol(), from.amount() - neededFromAmount);
        var newTo = new Currency(to.symbol(), to.amount() + requestedToAmount);
        return Stream.of(newFrom, newTo).collect(Collectors.toMap(Currency::symbol, currency -> currency));
    }

}
